/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals self, postMessage, importScripts, Map, Player, ROT, TECHS, ALLTECHS, ShipDesign, ShipDesignType */
/* globals HoldPositionShipOrder, HoldAndColoniseShipOrder, JumpShipOrder, JumpToAndColonisePlanetShipOrder */
/* globals ResearchOrder, CarryForwardOrder, ImproveColonyOrder, AddMarinesOrder, NullOrder */
/* globals BuyVehicleOrder, BuyStarBaseOrder */
/* exported onmessage */

if (!self.document) {
	importScripts('map.js?' + location.search);
	importScripts('ship.js?' + location.search);
	importScripts('shipdesign.js?' + location.search);
	importScripts('shiporder.js?' + location.search);
	importScripts('tech.js?' + location.search);
	importScripts('player.js?' + location.search);
	importScripts('order.js?' + location.search);
	importScripts('../rot/rot.min.js?' + location.search);

ShipDesign.init();

	var onmessage = function(message) {
		const datastylename = message.data.stylename;
		const databasemap = JSON.parse(message.data.baseMap);
		const datacapital = message.data.capital;
		const datatech = message.data.tech;
		const dataplayer = JSON.parse(message.data.player);

		const basemap = new Map(databasemap);
		const player = new Player(datastylename,datacapital,datatech);
		player.fromJSON(dataplayer,basemap);
		if (player.seed) {
			ROT.RNG.setSeed(player.seed); // Set ROT's seed on this computer worker thread
			for(let i=1; i<10; i++) {
				ROT.RNG.getUniform(); // Pull some numbers to get into a more random space
			}
		} else {
			ROT.RNG.setState(JSON.parse(player.ROTRNGState));
		}

		let orders = [new NullOrder()];
		let shiporders = [];

		if (player.active) {
			player.computeCommsGrid();
			moveShips(player);
			orders = spendMoney(player);

			for(let ship of player.ships) {
				shiporders.push(ship.order);
			}
		}

		// TODO: Ship movement also needs to fed back to the game
		let reply = {orders: JSON.stringify(orders), ships: JSON.stringify(shiporders), ROTRNGState: JSON.stringify(ROT.RNG.getState()) };
		postMessage(reply);
	};

	var moveShips = function(player) {
		for(let ship of player.ships) {
			if (ship.getJumpRange() > 0 && (ROT.RNG.getPercentage() > 40)) {
				const possibleHexes = [];
				for(let row=1; row<21; row++) {
					for(let column of "ABCDEFGHIJKLMNOP") {
						const name = column+row;
						if (Map.range(ship.location,name) <= ship.getJumpRange() && player.map.hex(name).onGrid) {
							possibleHexes.push(name);
						}
					}
				}
				const target = ROT.RNG.getItem(possibleHexes);
				if (ship.getColonyImprovement() > 0 && (ROT.RNG.getPercentage() > 40)) {
					ship.order = new JumpToAndColonisePlanetShipOrder(target);
				} else {
					ship.order = new JumpShipOrder(target);
				}
			} else if (ship.getColonyImprovement() > 0 && (ROT.RNG.getPercentage() > 40)) {
				ship.order = HoldAndColoniseShipOrder.SINGLETON;
			} else {
				ship.order = HoldPositionShipOrder.SINGLETON;
			}
		}
	};

	var spendMoney = function(player) {
		const orders = [];
		const colonies = [];
		const allHexesOnGrid = [];
		let money = player.production+player.carriedProduction;
		player.map.forEachHex(function(hex) {
			if (hex.owner == player.stylename && hex.level > 0) {
				colonies.push(hex.id);
			}
			if (hex.onGrid) {
				allHexesOnGrid.push(hex.id);
			}
		});
		// TODO Design new ships
		while(money > 0 && ROT.RNG.getPercentage() > 16) {
			const roll = ROT.RNG.getPercentage();
			if (roll < 34) {
				const amount = getRandom(1,money);
				const tech = ROT.RNG.getItem(ALLTECHS);
				orders.push(new ResearchOrder(tech,tech.index === player.techIndex ? amount*2 : amount));
				money -= amount;
			} else if (roll < 51) {
				const colony = ROT.RNG.getItem(colonies);
				if (player.map.hex(colony).level <= money/20) {
					orders.push(new ImproveColonyOrder(colony));
					money -= player.map.hex(colony).level*20;
				}
			} else if (roll < 66 && money > 9) {
				const colony = ROT.RNG.getItem(colonies);
				const companies = getRandom(1,Math.floor(money/10));
				orders.push(new AddMarinesOrder(colony,companies));
				money -= companies*10;
			} else if (roll < 95) {
				const selfPlayer = player;
				const selfMoney = money;
				const selfBuiltFactor = builtFactor;
				const colony = ROT.RNG.getItem(colonies.filter(colony => selfPlayer.map.hex(colony).level > 19));
				const possibleDesigns = ShipDesign.REGISTRY
					.filter(design => design.type !== ShipDesignType.StarBase)
					.filter(design => design.isTechSufficent({ power:	selfPlayer.researchLevel(TECHS.powerPlant),
						 jump:		selfPlayer.researchLevel(TECHS.jumpDrive),
						 man:		selfPlayer.researchLevel(TECHS.maneuverDrive),
						 size:		selfPlayer.researchLevel(TECHS.shipSize),
						 weapons:	selfPlayer.researchLevel(TECHS.weaponary),
						 defences:	selfPlayer.researchLevel(TECHS.shipDefences),
						 computer:	selfPlayer.researchLevel(TECHS.computers),
						 arch:		selfPlayer.researchLevel(TECHS.archaeology),
						 planet:	selfPlayer.researchLevel(TECHS.planetaryDefences)}))
					.filter(design => design.totalCost()*selfBuiltFactor(selfPlayer,design,orders) <= selfMoney);
				if (possibleDesigns.length > 0) {
					const design = ROT.RNG.getItem(possibleDesigns);
					const name = design.name+" "+getRandom(0,99999999);
					money -= design.totalCost()*builtFactor(player,design,orders);
					orders.push(new BuyVehicleOrder(design,name,colony));
				}
			} else {
				const location = ROT.RNG.getItem(allHexesOnGrid);
				const selfPlayer = player;
				const selfMoney = money;
				const selfBuiltFactor = builtFactor;
				const possibleDesigns = ShipDesign.REGISTRY
					.filter(design => design.type === ShipDesignType.StarBase)
					.filter(design => design.isTechSufficent({ power:	selfPlayer.researchLevel(TECHS.powerPlant),
						 jump:		selfPlayer.researchLevel(TECHS.jumpDrive),
						 man:		selfPlayer.researchLevel(TECHS.maneuverDrive),
						 size:		selfPlayer.researchLevel(TECHS.shipSize),
						 weapons:	selfPlayer.researchLevel(TECHS.weaponary),
						 defences:	selfPlayer.researchLevel(TECHS.shipDefences),
						 computer:	selfPlayer.researchLevel(TECHS.computers),
						 arch:		selfPlayer.researchLevel(TECHS.archaeology),
						 planet:	selfPlayer.researchLevel(TECHS.planetaryDefences)}))
					.filter(design => design.totalCost()*selfBuiltFactor(selfPlayer,design,orders) <= selfMoney);
				if (possibleDesigns.length > 0) {
					const design = ROT.RNG.getItem(possibleDesigns);
					const name = design.name+" "+getRandom(0,99999999);
					money -= design.totalCost()*builtFactor(player,design,orders);
					orders.push(new BuyStarBaseOrder(design,name,location));
				}
			}
		}
		orders.push(new CarryForwardOrder(money));
		return orders;
	};

	var builtFactor = function(player,design,orders) {
		let number = design.numberOfClassBuilt(player)+1;
		// Scan the orders looking for orders of this ship class - if so bump the number by one
		// TODO: this will get duplicates if an earlier order is cancelled - perhaps we should be logging the number ordered regardless of cancellations
		for(let order of orders) {
			if ((order instanceof BuyVehicleOrder) && (order.design === design)) { number++; }
		}
		// If the number is 1 AND this is a second or later ship we must double the price
		if (number === 1 && hasEverBuiltAShip(player,orders)) {
			return 2;
		}
		return 1;
	};

	var hasEverBuiltAShip = function(player,orders) {
		if (player.hasBuiltAShip) return true;
		for(let order of orders) {
			if (order instanceof BuyVehicleOrder) return true;
		}
		return false;

	};

	var getRandom = function(min,max) {
		return Math.floor(ROT.RNG.getUniform()*(max-min+1))+min;
	};
}