/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals ROT, Map, Player, Util, console */
/* globals  Order, NullOrder, PlayerOrders */
/* globals ShipOrderReport, ShipDesign, ShipDesigner */
/* globals ShipOrder, JumpShipOrder, JumpToAndColonisePlanetShipOrder, HoldAndColoniseShipOrder */
/* exported stylenames */

const InitialState = { version: "34", minimumVersion: "34", turn: 1 }; /* Initial Game State - version indicates which version of save data we create and can read */
const stylenames = ['player', 'alkari', 'auron', 'caffe', 'clanger', 'drahmini', 'gobb', 'iyanden', 'jerandi', 'kalli', 'kelle',
	'pican', 'regul', 'saggitt', 'strata', 'fleet', 'worati'];
const techs = [0, 8, 9, 6, 6, 5, 3, 1, 8, 8, 3, 9, 4, 5, 3, 2, 0];
var Game = {
	STYLENAMES: stylenames,
	FULLNAMES: {'player': "you",
				'alkari': "the Alkari",
				'auron':"the Aurons",
				'caffe':"the Caffenians",
				'clanger':"the Clangers",
				'drahmini':"the Drahmini",
				'gobb':"the Gobbannanean Principalities",
				'iyanden':"the Chroniomancy of Iyanden",
				'jerandi':"the Jerandi People's Republic",
				'kalli':"the Kallikanzaroi",
				'kelle':"the Kellesonians",
				'pican':"the Picans",
				'regul':"the Regul",
				'saggitt':"the Saggittarians",
				'strata':"Strata Inc.",
				'fleet':"the Fleet",
				'worati':"the Worati"
				},
	State: { version: undefined, baseMap: undefined, turn: undefined, playerProduction: undefined },
	_standardShipDesigns: [],
	_players: [],
	_stylenameToPlayerMapping: {},
	_COMPUTERURL: null,
	_computers: {},
	_orders:{},
	History: { baseMap: {} },
	ARTIFACT_DESIGN: null,
	_showMap: function(map) {
		map.forEachHex(function(hex) {
			const hexelement = document.getElementById(hex.id);
			if (hex.lastScanned) {
				const planetElement = hexelement.childNodes[3];
				planetElement.classList.remove("emptyspace", "starbase","planet");
				let style = "emptyspace";
				if (hex.starbase) style = "starbase";
				if (hex.planet) style = "planet";
				planetElement.classList.add(style);
				hexelement.classList.remove(...stylenames);
				if (hex.lastScanned && hex.owner) {
					hexelement.classList.add(hex.owner);
				}
				let text = "";
				if (hex.level) {
					text = "Class "+hex.level+" planet\n";
				}
				if (hex.owner) {
					text += "Owned by "+Game.FULLNAMES[hex.owner]+"\n";
				}
				hexelement.title = text + "Last report from Turn " + hex.lastScanned;
				if (hex.lastReport) {
					hexelement.title += "\n"+hex.lastReport;
				}
				hexelement.classList.remove("fogofwar");
			} else {
				hexelement.classList.add("fogofwar");
			}
		});
	},
	_generateNewGame: function() {
		Game.State = Object.assign({}, InitialState);
		Game.State.baseMap = new Map();
		Game.History = { baseMap: {}, version: Game.State.version };
		Game.State.baseMap.forEachHex(function(hex) {
			if (hex.planet) Game.History.baseMap[hex.id] = true;
		}); /* basemap shows where the planets are */

		// Now create the players
		Game._createPlayers(document.getElementById("favouredTech").selectedIndex);

		document.getElementById("welcome").classList.replace("visible", "hidden");
		document.getElementById("summary").classList.replace("hidden", "visible");
		PlayerOrders.setPlayer(Game._players[0]);
		PlayerOrders.setStandardShipDesigns(Game._standardShipDesigns);
		Game.State.turn = 0;
		Game._newTurn();
	},
	_createPlayers: function(playerTechIndex) {
		const seed = ROT.RNG.getSeed();
		const capital = {};
		Game.State.baseMap.forEachHex(hex => {
			if (hex.owner) {
				capital[hex.owner] = hex.id;
			}
		});
		Game._players = [];
		Game._stylenameToPlayerMapping = {};
		for(let i=0; i< Game.STYLENAMES.length; i++) {
			let player = new Player(Game.STYLENAMES[i],
				capital[Game.STYLENAMES[i]],
				(i === 0 ? playerTechIndex : techs[i]),
				seed);
			player.init(Game.State.baseMap,Game.State.turn);
			Game._players.push(player);
			Game._stylenameToPlayerMapping[Game.STYLENAMES[i]] = player;
		}
	},
	_newTurn: function() {
		Game.State.turn++;
		Game._updateHistory();
		Game._updateProduction();
		Game._saveGameState();
		Game._startTurn();
	},
	_startTurn: function() {
		Game._pokeComputerWorkers();
		Game._clearDisplayTables();
		Game._showResearchReports();
		Game._showColonyReports();
		ShipOrderReport.showShipReports(Game._players[0]);
		Game._showMap(Game._players[0].map);
		PlayerOrders.prepare();
		Util.replaceText("turnnumber", Game.State.turn);
		Game._orders[Game.STYLENAMES[0]] = null; // No orders from the player
	},
	_createComputerWorkers: function() {
		Game._computers = {};
		Game._orders = {};
		for(let i=1; i<Game.STYLENAMES.length; i++) {
			let worker = new Worker(Game._COMPUTERURL);
			worker.onmessage = Util.captureExceptionEventHandler(Game._computerHasOrders);
			worker.onerror = Util.captureExceptionEventHandler(Game._computerHasError);
			Game._computers[Game.STYLENAMES[i]] = worker;
		}
	},
	_pokeComputerWorkers: function() {
		for(let i=1; i<Game._players.length; i++) {
			let player = Game._players[i];
			let worker = Game._computers[player.stylename];
			Game._orders[player.stylename] = null;
			worker.postMessage({stylename: player.stylename,
								baseMap: JSON.stringify(Game.State.baseMap),
								capital: player.capital,
								tech:player.indexOfFavouredTech,
								player: JSON.stringify(player)
								});
			player.seed = 0; // So that the computer worker knows it's already seen the seed
		}
	},
	_clearDisplayTables: function() {
	// Clear existing rows after the templates
		for(let templaterow of document.querySelectorAll(".template")) {
			while (templaterow.nextSibling) {
				templaterow.parentNode.removeChild(templaterow.nextSibling);
			}
		}
	},
	_updateProduction: function() {
		const productionTotal = {};
		Game.State.baseMap.forEachHex(function(hex) {
			if (hex.owner && hex.level > 0) {
				productionTotal[hex.owner] = productionTotal[hex.owner] ? productionTotal[hex.owner]+hex.level*20 : hex.level*20;
			}
		});
		for(let player of Game._players) {
			player.production = productionTotal[player.stylename];
		}
	},
	// TODO Side effect: Also compute total production - this may (probably will) need to be refactored at some point
	_showColonyReports: function() {
		// Find the colony listing table
		let colonies = document.getElementById("colonies");
		Game._players[0].map.forEachHex(function(hex) {
			if (hex.owner == Game.STYLENAMES[0] && hex.level > 0) {
				const newrow = colonies.insertRow();
				const rowname = document.createElement("td"); rowname.append(hex.id);
				const report = document.createElement("td");
				report.append(hex.colonyReport ? hex.colonyReport : "Nothing to report");
				const marines = document.createElement("td"); marines.append(hex.marines);
				const level = document.createElement("td"); level.append(hex.level);
				const productionCell = document.createElement("td"); productionCell.append(hex.level * 20);
				newrow.append(rowname,report,marines,level,productionCell);
			}
		});
	},
	_showResearchReports: function() {
		Game._players[0].forEachTech( (tech, research) => {
			Util.replaceText(tech.domPrefix+"Points",research);
			let level = tech.level(research);
			Util.replaceText(tech.domPrefix+"Level",level);
			while(level > 1) {
				document.getElementById(tech.domPrefix+level).classList.remove("hidden");
				level--;
			}
		});
	},
	_updateHistory: function() {
		let sheet = {};
		Game.State.baseMap.forEachHex(function(hex) {
			if (hex.owner) sheet[hex.id] = hex.owner; /* Each turn has a map showing who owns the hex */
		});
		Game.History[Game.State.turn] = sheet;
		localStorage.domainOfTheAncientsHistory = JSON.stringify(Game.History);
	},
	init: function() {
		// Find the URL for the computer.js script
		const computerjs = document.getElementById("djv_domainOfTheAncients_computer-js");
		Game._COMPUTERURL = computerjs.src;

		ShipDesign.init();
		for(let shipdesign of ShipDesign.REGISTRY) { // Only valid because the players haven't been created or loaded yet!
			Game._standardShipDesigns.push(shipdesign);
		}
		Game.ARTIFACT_DESIGN = ShipDesign.REGISTRY[0];

		Game._initGameButtons();
		Game._initTwisties();
		PlayerOrders.init();
		Game._createComputerWorkers();
		ShipDesigner.init();
	},
	_initGameButtons: function() {
		document.getElementById("generate").onclick = Util.captureExceptionFunction(Game._generateNewGame);
		if (localStorage.domainOfTheAncients && localStorage.domainOfTheAncientsVersion && localStorage.domainOfTheAncientsSaveDate) {
			Game._showLoadButton();
			let v;
			try {
				v = JSON.parse(localStorage.domainOfTheAncientsVersion);
			} catch(_e) {
				v = null; // Clearly no version data if we can't parse it!
			}
				if (v && v.game && (parseInt(v.game) >= InitialState.minimumVersion)) {
					document.getElementById("resume").disabled = false;
					document.getElementById("wrongVersion").classList.replace("visible","hidden");
					document.getElementById("resume").onclick = Util.captureExceptionFunction(Game._load);
				} else {
					document.getElementById("resume").disabled = true;
					if (v && v.plugin) {
						Util.replaceText("incompatibleVersionNumber",v.plugin);
					}
					document.getElementById("wrongVersion").classList.replace("hidden","visible");
				}
		} else {
			document.getElementById("resume").disabled = true;
		}
		if (localStorage.domainOfTheAncientsHistory) {
			// Nothing yet!
		} else {
			document.getElementById("replay").disabled = true;
		}

		document.getElementById("nextTurn").onclick = Util.captureExceptionFunction(Game._humanHasOrders);
		document.getElementById("processorderread").onclick = Util.captureExceptionFunction(Game._completeTurn);
	},
	_initTwisties: function() {
		// Look for the twisties and set them up
		for(let button of document.querySelectorAll("h4 button")) {
			// Look for the child svg's 'we are interested in
			let svg1 = null, svg2 = null, followingdiv = null;
			for(let child of button.childNodes) {
				if (child.nodeName == "svg") {
					if (svg1 === null) {
						svg1 = child;
					} else {
						svg2 = child;
					}
				}
			}
			// And then the uncle's' (i.e. parent's sibling) div
			let sibling = button.parentNode.nextSibling;
			while(followingdiv === null && sibling) {
				if (sibling.nodeName === "DIV") {
					followingdiv = sibling;
				} else {
					sibling = sibling.nextSibling;
				}
			}

			if (svg1 && svg2 && followingdiv) {
				button.onclick = Util.captureExceptionFunction(Util.toggleFields,[svg1,svg2,followingdiv]);
			}
		}
	},
	_computerHasError: function(event) {
		Util.reportException(event);
		const worker = event.currentTarget;
		for(let stylename of Object.keys(Game._computers)) {
			if (Game._computers[stylename] === worker) {
				Game._orders[stylename] = [new NullOrder()];
				console.log(`It was the worker for ${stylename} that encountered ${event}`);
				break;
			}
		}
	},
	_computerHasOrders: function(event) {
		const worker = event.currentTarget;
		for(let stylename of Object.keys(Game._computers)) {
			if (Game._computers[stylename] === worker) {
				const spend = JSON.parse(event.data.orders);
				Game._orders[stylename] = [];
				for(let orderdata of spend) {
					Game._orders[stylename].push(Order.fromJSON(orderdata));
				}
				const shiporders = JSON.parse(event.data.ships);
				for(let i=0; i<shiporders.length; i++) {
					Game._stylenameToPlayerMapping[stylename].ships[i].order = ShipOrder.fromJSON(shiporders[i]);
				}
				Game._stylenameToPlayerMapping[stylename].ROTRNGState = event.data.ROTRNGState; // Ready to resend next time
				break;
			}
		}
		if (!Game._processIfEverybodyReady()) {
			Game._updateAwaitingList();
		}
	},
	_humanHasOrders: function() {
		Game._orders[Game.STYLENAMES[0]] = PlayerOrders.getOrders();
		if (!Game._processIfEverybodyReady()) {
			// Need to wait for more orders from computer players, so display the status page
			document.getElementById("summary").classList.add("hidden");
			document.getElementById("awaitingOrders").classList.remove("hidden");
			Game._updateAwaitingList();
		}
	},
	_updateAwaitingList: function() {
		if (!document.getElementById("awaitingOrders").classList.contains("hidden")) {
			const list = document.getElementById("awaitingOrdersFromList");
			Util.emptyList(list);
			for(let stylename of Object.keys(Game._computers)) {
				if (!Game._orders[stylename]) {
					const term = document.createElement("dt");
					const def = document.createElement("dd");
					Util.replaceText(term,stylename);
					Util.replaceText(def,"");
					list.append(term,def);
				}
			}
		}
	},
	_processIfEverybodyReady: function() {
		let ready = true;
		for(let i=0; i<Game._players.length; i++) {
			if (!Game._orders[Game._players[i].stylename]) {
				ready = false; break;
			}
		}
		if (ready) {
			Game._processTurn();
			return true;
		} else {
			return false;
		}
	},
	_processTurn: function() {
		document.getElementById("awaitingOrders").classList.add("hidden");
		const summary = document.getElementById("summary");
		const proc = document.getElementById("processOrder");
		const updates = [];
		const shipsToBeDestroyed = [];

		// Ship orders
		// Step 1. create a new map and place all ships in their new locations
		const workingMap = new Map(Game.State.baseMap);
		for(let player of Game._players) {
			for(let ship of player.ships) {
				if (ship.order instanceof JumpShipOrder || ship.order instanceof JumpToAndColonisePlanetShipOrder) {
					ship.location = ship.order.target;
				}
				workingMap.addObject(ship.location, {ship: ship, owner:player});
			}
		}

		// Step 2. Resolve combat
		// TODO !!

		// Step 3. Resolve ship orders that remain after movement
		workingMap.forEachHex(hex => {
			// Step 3a. Work out what is in the hex
			const bystyle = {}; // each element (when present) has a sparse array showing how many ships of that size are present
			for(let object of hex.objects) {
				if (object.ship && object.owner) {
					bystyle[object.owner] = bystyle[object.owner] ? bystyle[object.owner] : [];
					bystyle[object.owner][object.ship.size] = bystyle[object.owner][object.ship.size] ? bystyle[object.owner][object.ship.size]++ : 1;
				}
			}

			// Step 3b. Give the colony (if any) a report
			if (hex.owner && hex.level) {
				let otherowners = Object.assign([],bystyle); delete otherowners[hex.owner];
				let report = "";
				for(let otherowner of otherowners.keys()) {
					for(let size=1; size<otherowners[otherowner].length; size++) {
						if (otherowners[otherowner][size]) {
							report += "There " +
								(otherowners[otherowner][size]==1 ? "is a" : "are "+otherowners[otherowner][size]) +
								" size "+size+" ship" +
								(size==1 ? "" : "s") +
								" belonging to " +
								Game.FULLNAMES[otherowner] +
								" here. ";
						}
					}
				}
				Game._stylenameToPlayerMapping[hex.owner].map.setColonyReport(hex.id,report);
				Game._stylenameToPlayerMapping[hex.owner].map.setLastScanned(hex.id,Game.State.turn,report);
			}

			// Step 3c. And then for each ship determine what it does
			for(let object of hex.objects) {
				if (object.ship && object.owner) {
					const ship = object.ship;
					const owner = object.owner;
					ship.report = "";

					// Scanning of the hex
					const wasUnscanned = (owner.map.hex(hex.id).lastScanned ? false : true); // Must look at the owner's map for this detail'
					if (hex.planet && hex.owner) {
						ship.report += `${ship.location} is a class ${hex.level} colony belonging to ${Game.FULLNAMES[hex.owner]}. `;
					} else if (hex.planet) {
						ship.report += `${ship.location} is an uncolonised world. `;
					}

					let informHuman = (owner == Game._players[0] && wasUnscanned);

					let otherowners = Object.assign([],bystyle); delete otherowners[owner];
					for(let otherowner of otherowners.keys()) {
						for(let size=1; size<otherowners[otherowner].length; size++) {
							if (otherowners[otherowner][size]) {
								ship.report += "There " +
									(otherowners[otherowner][size]==1 ? "is a" : "are "+otherowners[otherowner][size]) +
									" size "+size+" ship" +
									(size==1 ? "" : "s") +
									" belonging to " +
									Game.FULLNAMES[otherowner] +
									" here. ";
							}
						}
						informHuman = (owner == Game._players[0]); // We assume the player will always want to know about ships in the same hex as one of theirs (at least for now)
					}

					// Update the hex report before the ship colonises
					owner.map.setLastScanned(hex.id,Game.State.turn,ship.report); // If a player has multiple ships we will set this multiple times (to the same value)

					// Colonisation (if ordered to do so)
					if (ship.order instanceof HoldAndColoniseShipOrder || ship.order instanceof JumpToAndColonisePlanetShipOrder) {
						if (hex.planet && hex.owner !== owner.stylename && hex.level > 0) {
							ship.report += `Unable to colonise ${hex.id}. `;
							informHuman = (owner == Game._players[0]);
						} else if (hex.planet && hex.owner !== owner.stylename && hex.level == 0) {
							ship.report += `There is a star base belonging to ${Game.FULLNAMES[hex.owner]} here. Unable to colonise. `;
							informHuman = (owner == Game._players[0]);
						} else if (hex.planet && hex.owner == owner.stylename && hex.level < 5) {
							ship.report += `Colonisation of ${hex.id} has been achieved adding ${ship.getColonyImprovement()} to the existing colony. `;
							informHuman = (owner == Game._players[0]);
							Game.State.baseMap.colonise(ship.location,owner.stylename,ship.getColonyImprovement());
							hex.level += ship.getColonyImprovement();
							shipsToBeDestroyed.push(object);
						} else if (hex.planet && hex.owner == owner.stylename) {
							ship.report += `Colonisation of ${hex.id} cannot be made as the colony is already at level ${hex.level}. `;
							informHuman = (owner == Game._players[0]);
						} else if (hex.planet) {
							ship.report += `Colonisation of ${hex.id} has been achieved. `;
							informHuman = (owner == Game._players[0]);
							Game.State.baseMap.colonise(hex.id,owner.stylename,ship.getColonyImprovement());
							shipsToBeDestroyed.push(object);
						}
					}

					// Inform the human if necessary
					if (informHuman && ship.hasAReport()) {
						updates.push(ship.name+" reports that "+ship.report);
					}
				}
			}
		});

		// Step 4. Remove those ship no longer in existence
		for(let object of shipsToBeDestroyed) {
			object.owner.destroyShip(object.ship);
		}

		// Step 6. Apply non-ship orders
		for(let i=0; i<Game._players.length; i++) {
			const stylename = Game._players[i].stylename;
			for(let order of Game._orders[stylename]) {
				let update = order.apply(Game._players[i],Game.State.baseMap,Game.State.turn,i===0);
				if (update) {
					// There's an update and since it's for the human we want to add it to the list of moves
					updates.push(update);
				}
			}
		}

		// Step 7 Let the human know of any interesting events (if any)
		if (updates.length > 0) {
			Game._showMap(Game._players[0].map); // To show the ship discoveries
			summary.classList.add("hidden");
			proc.classList.remove("hidden");
			const list = document.getElementById("messages");
			list.replaceChildren();
			for(let update of updates) {
				const element = document.createElement("li");
				element.append(update);
				list.append(element);
			}
		} else {
			Game._completeTurn(); // Just complete the turn
		}
	},
	_completeTurn: function() {
		let summary = document.getElementById("summary");
		let proc = document.getElementById("processOrder");
		proc.classList.add("hidden");
		summary.classList.remove("hidden");
		Game._newTurn();
	},
	_showLoadButton: function() {
		if (localStorage.domainOfTheAncientsSaveDate) {
			const saved = new Date(localStorage.domainOfTheAncientsSaveDate);
			const nextUpdate = Util.next_time_diff(saved);
			if (localStorage.domainOfTheAncientsTurnNumber) Util.replaceText("savegameturn",localStorage.domainOfTheAncientsTurnNumber);
			Util.replaceText("savegamedate",Util.time_diff(saved));
			// Is the welcome screen hidden?
			const hiddenWelcome = document.getElementById("welcome").classList.contains("hidden");
			if (nextUpdate && !hiddenWelcome) {
				window.setTimeout(Util.captureExceptionFunction(Game._showLoadButton, nextUpdate*1000));
			}
		}
	},
	_saveGameState: function() {
		let data = {};
		data.turn = Game.State.turn;
		data.baseMap = Game.State.baseMap;
		data.playerTech = Game._players[0].techIndex;
		data.players = [];
		for(let p of Game._players) {
			data.players.push(p);
		}
		data.built = [];
		for(let design of Game._standardShipDesigns) {
			data.built.push(design._built);
		}
		localStorage.domainOfTheAncientsTurnNumber = Game.State.turn;
		localStorage.domainOfTheAncientsSaveDate = (new Date()).toISOString();
		let version = {game: Game.State.version, plugin:"3.176.0"};
		localStorage.domainOfTheAncientsVersion = JSON.stringify(version);
		localStorage.domainOfTheAncients = JSON.stringify(data);
	},
	_load: function() {
		const data = JSON.parse(localStorage.domainOfTheAncients);
		Game.State.turn = data.turn;
		Game.State.baseMap = new Map(data.baseMap);
		Game._createPlayers(data.playerTech);
		for(let i=0; i<Game._players.length; i++) {
			Game._players[i].fromJSON(data.players[i],Game.State.baseMap);
			Game._stylenameToPlayerMapping[Game._players[i].stylename] = Game._players[i];
		}

		let i=0;
		for(let design of Game._standardShipDesigns) {
			design._built = data.built[i++];
		}

		PlayerOrders.setPlayer(Game._players[0]);
		PlayerOrders.setStandardShipDesigns(Game._standardShipDesigns);

		const v = JSON.parse(localStorage.domainOfTheAncientsVersion);
		Game.State.version = v.game;

		document.getElementById("welcome").classList.replace("visible", "hidden");
		document.getElementById("summary").classList.replace("hidden", "visible");
		Game._startTurn();
	},
};

window.onload = function() {
	Util.captureException(Game.init);
};
