/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals module, exports, ShipDesign, ShipOrder, HoldPositionShipOrder */
/* exported Ship */
class Ship {
	constructor(name,design,startLocation) {
		this._name = name;
		this._design = design;
		this._location = startLocation;
		this._order = HoldPositionShipOrder.SINGLETON;
		this._report = "Nothing to report";
	}
	get name() { return this._name; }
	get location() { return this._location; }
	set location(location) { this._location = location; }
	get report() { return this._report;  }
	set report(report) { this._report = report; }
	resetReport() { this._report = ""; }
	hasAReport() { return this._report !== ""; }
	get order() { return this._order; }
	set order(order) { this._order = order; }
	resetOrder() { this._order = HoldPositionShipOrder.SINGLETON; }
	getJumpRange() {
		// TODO account for damage
		return this._design.getJumpRange();
	}
	getColonyImprovement() {
		return this._design.getColonyImprovement(); // TODO Handle ship damage
	}
	get size() {
		return this._design.totalSize();
	}
	toJSON() {
		const data = {name: this._name, designId: this._design.id, loc: this._location, order: this._order, report: this._report };
		return data;
	}
}
Ship.fromJSON = function(data) {
	const ship =  new Ship(data.name, ShipDesign.REGISTRY[data.designId],data.loc);
	ship._order = ShipOrder.fromJSON(data.order);
	ship._report = data.report;
	return ship;
};

//For testing via QUnit
if (typeof module !== "undefined" && module.exports) {
    exports.Ship = Ship;
}